
#ifndef _skoplib_h_
#define _skoplib_h_

// General type definitions

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <memory.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
//#include <crtdbg.h>

// Error codes
#define		SUCCESS				1
#define		E_OUTOFMEMORY1		0xF001
#define		E_INVALIDFORMAT		0xF002
#define		E_FILEERROR			0xF003
//#define		E_UNEXPECTED		0xF004


// Min sizes
#define		ALPHABET_SIZE	20
#define		MAX_LEN			6000
#define		MAX_BANK		500
#define		MAX_KERNELS		4096 // 10240//1024
#define		MAX_PAIR_FAMILY	500 // 2000
#define		MAX_TRIPLET_FAMILY	400
#define		POOL_SIZE		72000
#define		MAX_STR			1500
#define		MAX_KERNEL_IN_ALI	200
#define     STRING_STANDR_LEN	255 // Nat

// Alignment path codes
#define DIR_DIAG                        0x0
#define DIR_LEFT                        0x1
#define DIR_TOP                         0x2
#define DIR_EMPTY                       0x3
#define DIR_TERMINATOR					0x4
#define DIR_LEFTOPEN					0x20
#define DIR_TOPOPEN						0x40
#define MISMATCH_POS            ':'
#define MISMATCH_NEG            ' '
#define MATCH                           '|'
#define DEL_CHAR                        '.'
#define DELTA                           1
#define PREC                            10E-4

#define	TOP_OPENED						0x1
#define	LEFT_OPENED						0x2


#define		GAP_SYMBOL		'.'

typedef int  		score_type;
typedef long		coord_type;
typedef unsigned	ptr_type;

typedef long*		long_pointer;

// Sequence
typedef struct _sequence {
	char*		name;	// name of the sequence
	char*		aas;	// string of aminoacid indices
	unsigned	length;	
} sequence;

// Sequence bank
// Multiple alignments in Clustal format
typedef struct _malign_bank {
	char*		name; // title of the clustal file
	char**		aligned_seq_strings; // strings of aligned sequences "FKL..AV.WP"

	sequence*	seqs;	// array of sequences
	unsigned	size;	// count of sequencxes in the bank
} malign_bank;


//
// Sequence bank in FASTA format
typedef struct _fasta_bank {
	char*		name;	// filename of fatsa bank
	sequence*	seqs;	// array of sequences
	unsigned	size;	// count of sequencxes in the bank		
} fasta_bank;

// pair-wise aligment
typedef struct _pair_aligment {
	char*		aligned_seq1;
	char*		aligned_seq2;
} pair_aligment;


typedef struct _aln_path {
	int			path[MAX_LEN];
	unsigned	length;
} aln_path;
// 
// Sequence bank in pairwise format
typedef struct _pairfull_bank {
	char*			name; // title of bank
	sequence*		seqs; // array of sequences
	unsigned		size; // count of sequences in the bank
	aln_path**		paths_GS;//[MAX_LEN]; // but used only top rigth angle of the matrix
	//pair_aligment*	pair_aln; // all pair_aligments
	//unsigned		pair_num; // number of pair_aligments
} pairfull_bank;

//--------------------------------------------------------------
// Kernels

// kernel definition
typedef struct _kernel {
	unsigned	id;
	coord_type	x, y, length;
	score_type	score;
} kernel;

typedef enum _kernel_method {
	kernel_method_matches,
	kernel_method_pair_primers,
	kernel_method_triplet_primers,
} kernel_method;

// parameters for building set of kernels
typedef struct _kernel_params {
	kernel_method	method;
	coord_type	min_length;
	score_type	min_score;
	score_type	max_hole;
	score_type	allowed_gap;
	score_type	pair_cutoff;
	score_type	triplet_cutoff;
	unsigned	ker_method; // debug nat

} kernel_params;

// array of N * 20 elements (like a profile)
typedef	coord_type	coord_array20[20];
typedef coord_array20* profile;

// Pair primers
typedef struct _kernel_pair_primer_data {
	long_pointer	pair_families[400];	// data for global initialization
	long_pointer	pair_starts[400];	// data for per-query initialization
	
} kernel_pair_primer_data;

// Triplet primers
typedef struct _kernel_triplet_primer_data {
	long_pointer	triplet_families[8000];	// data for global initialization
	long_pointer	triplet_starts[8000];	// data for per-query initialization
	
} kernel_triplet_primer_data;

// initialization data for the first sequence
typedef struct _kernel_compute_data {
	sequence*	seq1;
	sequence*	seq2;
	union {
		profile		match_starts;
		kernel_pair_primer_data	pair_primer_data;
		kernel_triplet_primer_data	triplet_primer_data;
	};
	char*		mask_matrix;
	kernel		kernels[MAX_KERNELS*20];
	unsigned	kernel_count;
	unsigned	coverage; // DEBUG purpose
	unsigned	total_length; // DEBUG purpose
	unsigned	used_length;  // DEBUG purpose
	unsigned	strong_coverage;  // DEBUG purpose
	kernel_params* params;
} kernel_compute_data;
//--------------------------------------------------------------
// Islands
typedef struct _islands_stat {
    unsigned	isl_index;

	unsigned	isl_len;
	unsigned	isl_correct_count;

	double		isl_sim; // = isl_correct_count / isl_len
	double		isl_conf; // = isl_correct_count / ? - not use yet

	score_type	isl_score;
	score_type	max_score; // max weight of island's part

	char		have_kern;
	unsigned	kernel_count;

	score_type	kern_sum_weight; // kernels in island
	unsigned	kern_sum_len;

	score_type	kern_max_weight; // kernels in island
	unsigned	kern_max_len;

	score_type	kern_full_weight; // full weight of kernel intersecting with island
	unsigned	kern_full_len;

} islands_stat;



//--------------------------------------------------------------
// Smith-waterman alignments

typedef struct _nw_node {
	score_type	best_score;
	score_type	jump_score;
	char		best_link;
	char		jump_link;
} nw_node;

typedef struct _sw_params {
	score_type	GOP;
	score_type	GEP;
} sw_params;

typedef struct _sw_link {
	char	link;
	char	open_flag;
} sw_link;

typedef struct _sw_cell {
	score_type	left;
	score_type	diag;
	score_type	top;
	score_type	jump;
} sw_cell;

typedef struct _sw_compute_data {
	sequence*	seq1;
	sequence*	seq2;
	sw_cell		line1[MAX_LEN];
	sw_cell		line2[MAX_LEN];
	score_type	max_weight;
	int			path[MAX_LEN]; // path(x) -> y 

} sw_compute_data;

typedef struct _sparse_params {
	score_type	KCP;
	score_type	GEP;
} sparse_params;

//------------------------------------------------------------
// Multiple matrix sw-alignments

//------------------------------------------------------------
// External global data declarations
extern const char*	amino_acids;
extern char		char_to_index[255];
extern score_type subst_matrix[ALPHABET_SIZE][ALPHABET_SIZE];
extern score_type  seq_matrix[ALPHABET_SIZE * MAX_LEN];

extern kernel_compute_data G_kernel_compute_data;
extern sw_compute_data G_sw_compute_data;
extern FILE* G_debug_dump;

extern int G_positive_improve;
extern int G_negative_improve;

int	 matrix_read_from_file(char* _filename);
int matrix2_read_from_file(char* _filename);

void init_char_to_index();

void init_seq_matrix( sequence * seq1); // change seq_matrix 

int  read_bank_from_file(char* _filename, malign_bank* _bank, unsigned _z_cutoff);
int  malign_bank_read_from_file(char* _filename, malign_bank* _bank);
int  malign_bank_read_from_FSSPfile(char* _filename, malign_bank* _bank, unsigned _cutoff_z);
void malign_bank_release(malign_bank* _bank);

int  delete_seq_from_malign_bank(malign_bank* _bank, unsigned index);

int	merge_two_banks(malign_bank* _bank1, malign_bank* _bank2);
int save_bank_as_clustal(malign_bank* _bank, char* _filename);

int read_bank_as_fasta_from_file(char* _filename, fasta_bank* _bank);
int fasta_bank_read_from_file(char* _filename, fasta_bank* _bank);
int  add_seq_in_fasta_bank(fasta_bank* _bank, sequence* _seq); 
int clustal_to_fasta(malign_bank* _bank_clustal, fasta_bank* _bank_fasta);

int read_fasta_resalign(char * file_name, char* align_str_1, char* align_str_2);
int read_blast_resalign(char * file_name, char* align_str_1, char* align_str_2);

void fasta_bank_release(fasta_bank* _bank);
void seq_free(sequence* _seq);

int is_clustal(char* _file_name);
int is_clustal_or_fssp(char* _file_name);

int read_seq_file (char* _file_name, sequence* _seq);
int read_seq_fasta (char* _file_name, sequence* _seq);
int read_seq_PDB (char* _file_name, sequence* _seq, const char strand);

int read_seq_from_pdb (char* seq_name, char  strand, sequence* seq_seq);
void print_seq_as_fasta(char* file_name, sequence* seq, char* open_file_mode);

char convert_am_ac_to_1(char* s);
char convert_3am_to_1index(char* s);
char convert_3am_to_1(char* s);

int init_compute_kernels(sequence _seq1);
int compute_kernels(sequence* _seq1, sequence* _seq2, kernel_params* _params);
void kernel_compute_data_release(kernel_compute_data* _data);
void kernel_init_data_release(kernel_compute_data* _data);

int debug_print_kernels(char* _file_name, kernel* _kernels, unsigned _kernel_count, char* _seqname1, char* _seqname2, unsigned _len1, unsigned _len2, float _id);
void print_pair_primer_families(char* _file_name);
void print_pair_starts(char* _file_name);

unsigned reget_kernels(kernel* _kernels, unsigned _count, unsigned _required);
unsigned reget_point_count(kernel* _kernels, unsigned _count);

void mysort(kernel* _kernels, unsigned _count);
int  kernel_compare(const void * arg1, const void * arg2);

unsigned get_kern_count(kernel* kernels, unsigned kernel_count, score_type weight); 
int get_bank_Nrel(fasta_bank * _bank, char* seq_name);
void get_scop_code( char* seq_name, char* code);

int read_FASTA_search_file(char* res_file_name, unsigned Nrel, char* query_name,
						   float* sort,float* n1,float* n2, unsigned* tot);

int read_BLAST_search_file(char* res_file_name, unsigned Nrel, char* query_name,
						   float* sort,float* n1,float* n2, unsigned* tot);

float estimate_sort_skop(unsigned Nrel,unsigned numb,char* list);
float estimate_n2(unsigned Nrel,unsigned numb,char* list);
float estimate_n1(unsigned Nrel,unsigned numb,char* list);

//===========================================================
int init_pair_primer_families(kernel_compute_data* _data);
void pair_primer_families_release(kernel_compute_data* _data);
int init_compute_pair_primer_kernels(sequence _seq1);
int compute_pair_primer_kernels(sequence* _seq1, sequence* _seq2, kernel_params* _params);
void pair_primer_kernels_release(kernel_compute_data* _data);

//===========================================================
int init_triplet_primer_families(kernel_compute_data* _data);
void triplet_primer_families_release(kernel_compute_data* _data);
int init_compute_triplet_primer_kernels(sequence _seq1);
void triplet_primer_kernels_release(kernel_compute_data* _data);
int compute_triplet_primer_kernels(sequence* _seq1, sequence* _seq2, kernel_params* _params);

//===========================================================
int compute_sw_alignment(sequence* _seq1, sequence* _seq2, sw_params* _params);
int compute_sw_alignment_cut(sequence*	_seq1,sequence*	_seq2,sw_params*	_params);

int compute_sw_alignment_cut_seqsubst(	// need activated seq_matrix
	sequence*	_seq1,					// with _seq1
	sequence*	_seq2,
	sw_params*	_params
);
int compute_sw_alignment_seqsubst(	// need activated seq_matrix
	sequence*	_seq1,					// with _seq1
	sequence*	_seq2,
	sw_params*	_params
);


int compute_sw_alignment_dblgap(sequence*	_seq1, sequence* _seq2, sw_params* _params);

void sw_data_release(sw_compute_data* _data);
int format_sw_alignment_output(sw_compute_data* _data, char* _align1, char* _align2);

void print_sw_params_in_rezult( FILE* res_file,	sw_params*_params, char* matrix_name);

//===========================================================
int format_alignment_output(sequence* _seq1, sequence* _seq2, int* _path, char* _align1, char* _align2);
int print_alignment_decorated(sequence* _seq1, sequence* _seq2, int* _path_GS, int* _path_skop, kernel* _kernels, unsigned _count, unsigned* _mask, FILE* _f_out);
int print_alignment_ext(sequence* _seq1, sequence* _seq2, int* _path_GS, int* _path_skop, kernel* _kernels, unsigned _count, unsigned* _mask, FILE* _f_out);

void print_alignment(char* name1,char* name2,
					  char* align_str_1,char* align_str_2,
					  float ID, score_type weight,
					  FILE * f);

//===========================================================
// alignment representation & comparision
void format_alignment_path(char* _align1, char* _align2, int* _path, unsigned* _len);
float calc_alnide(int* _path_GS, int* _path_sample, unsigned _len_GS, unsigned* _match);
float calc_kernel_alnide(int* _path_GS, int* _path_sample, unsigned _len_GS, unsigned* _mask, unsigned _length2);
float calc_ide(char* _align1, char* _align2);
unsigned calc_match_count(int* _path, unsigned _length);

unsigned* create_mask(sequence* _seq1, sequence* _seq2, kernel* _kernels, unsigned _kernel_count);
void process_alignment(sequence* _seq1, sequence* _seq2, int* _path, int* _path_GS, unsigned* _mask, unsigned* _used_count, unsigned* _GS_count, unsigned* _corr_count);

unsigned get_isl_count(int * _path, unsigned len);
int island_compare_with_kernels(sequence* _seq1, sequence* _seq2, 
								int* _path, int* _pathGS, 
								unsigned* _mask,
								kernel* _kernels, unsigned _kernel_count,
								islands_stat* isl_stat);
int island_max_stat_and_print(char* file_name, char* family_name,
							 sequence* _seq1, sequence* _seq2, 
								int* _path, int* _pathGS, 								
								unsigned* _mask,
								double id_,double _acc, double _conf,
								kernel* _kernels, unsigned _kernel_count
								);
void print_islands_statistic (char* file_name,char* family_name,
							  islands_stat* isl_stat, unsigned isl_count, 
							  sequence* _seq1, sequence* _seq2, 
							  double id_, double _acc, double _conf );
void print_max_islands_statistic (char* file_name,char* family_name,
							  islands_stat* isl_stat, unsigned isl_count, 
							  sequence* _seq1, sequence* _seq2, 
							  double id_,double _acc, double _conf );

int island_compare_and_print(char* file_name, char* family_name,
							 sequence* _seq1, sequence* _seq2, 
								int* _path, int* _pathGS, 								
								unsigned* _mask,
								float id_, double _acc, double _conf,
								kernel* _kernels, unsigned _kernel_count
								);
int is_in_kernel(coord_type _x, coord_type _y, 
					  kernel* _kernels, unsigned _kernel_count
					  );

unsigned ker_count_used_in_aln(kernel* kernels, unsigned kernel_count,
						       score_type weight, // we are interested only in kernels.score >= weight  
							   //unsigned* _ker_count, // number of such kernels
							   int* path,
							   unsigned path_len,
							   unsigned* mask, unsigned len2);


int compare_with_pdb( sequence *seq, sw_params* _sw_params, 
					  int* pathres, sequence *seqPDB, FILE* f_out );
void compare_with_pdb_revers( sequence *seq, sw_params* _sw_params, 
					  int* pathres2, sequence *seqPDB, FILE* f_out );
int fulfill_full_seqbank( malign_bank* short_bank, pairfull_bank* full_bank,
						 sw_params* _params,
						 char* pdb_file_path);

void print_full_seqbank(pairfull_bank* full_bank, char* file_name);
void debug_print_path(int* path, unsigned len, char* file_name);

void free_full_seqbank(pairfull_bank* _bank);


int save_aln_asFASTA(int* path, sequence *seq1, sequence *seq2, char del_symbol);
double calculate_distance (sequence *seq1, sequence *seq2, int* path, double* dists,
						   char* pdb_file_path);

void make_pdb_filename(char* seq_name, char* res_file_name, char* stand);

double calc_dist_alnide(int* path_GS, double* dist_GS, int* path, double* dist, unsigned _len_GS);
double calc_dist_kernel_alnide(int* path_GS, double* dist_GS, 
							  int* path,    double* dist,
							  unsigned _len_GS, unsigned* _mask, unsigned _length2);
double calc_dist_footstep_alnide(int* path_GS, double* dist_GS, 
								 int* path,    double* dist, 
								 unsigned _len_GS,
								 double Dist);
double calc_dist_footstep_kernel_alnide(int* path_GS, double* dist_GS, 
										int* path,    double* dist,
										unsigned _len_GS, unsigned* _mask, unsigned _length2,
										double Dist);
double calc_dist_footstep_conf(
						int* path_GS, double* dist_GS, 
						int* path,    double* dist, 
						unsigned _len,
						double Dist);
double calc_dist_footstep_kernel_conf(
							   int* path_GS, double* dist_GS, 
							   int* path,    double* dist,
							   unsigned _len, unsigned* _mask, unsigned _length2,
							   double Dist);



void standardize_del_simbols(char* aln_str, char old_del);
int extend_algn_on_full_seqs(char* align_str_1, char* align_str_2,
							 sequence*	seq1  , sequence*	seq2,
							 int* path);

void convert_to_fasta( sequence * seq);
void convert_to_fasta2( sequence * seq); // without realloc
void copy_seq(sequence * init_seq, sequence * res_seq);
void get_mem_for_seq(sequence * seq);

score_type Calc_weight (char* aln_seq1, char* aln_seq2,
						char* Matrix_file, score_type GOP, score_type GEP,
						char* Init_Matrix);
score_type Calc_weight_subst2 (char* aln_seq1, char* aln_seq2,
							   score_type GOP,	score_type GEP);

#endif //_skoplib_h_
